/***************************************************************************
 *   QGame++                                                               *
 *   =======                                                               *
 *   A Quantum Gate And Measurement Emulator.                              *
 *                                                                         *
 *   Copyright (C) 2003/04 by Manuel Nickschas                             *
 *   <qgame-devel@nickschas.de>                                            *
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program; if not, write to the                         *
 *   Free Software Foundation, Inc.,                                       *
 *   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.             *
 ***************************************************************************/

#include "client.h"
#include <string>
#include <iostream>
#include <fstream>
#include <sstream>

using namespace std;
using namespace qgame;

/* This is a command line client for QGame++. This is more or less a
 * proof-of-concept sort of thing, which provides the basic functions
 * and not more. It is also (almost) not documented at this time, and
 * the documentation contained in the QGame++ package focuses on 
 * libqgame++ only, which is the main contents of the package rather than
 * this client.
 * If you wish further development of the client, or have any questions, 
 * please drop me a message.
 */

int clientMain(int argc, char **argv) {
  try {
    Settings settings;  
    getOptions(argc, argv, settings);
    QProgram prg; QProgramSettings prgSets;
    if(!strcmp(settings.filename,"-")) {
      prgSets = prg.read(cin);
    } else {
      ifstream fin;
      fin.open(settings.filename,ios::in);
      if(!fin.is_open()) throw(Error(E_IO));
      //readSettings(settings,fin); cerr << "\nvorher: " << fin.tellg();
      //fin.seekg(0,ios::beg); cerr << "\nnachher: " << fin.tellg();
      prgSets = prg.read(fin);
      fin.close();
    }
    if(settings.numQb < 0) settings.numQb = prgSets.numQb;
    if(settings.cases.size()<=0) settings.cases = prgSets.cases;
    if(settings.threshold<0) settings.threshold = prgSets.threshold;
    if(settings.finalQb.size()<=0) settings.finalQb = prgSets.finalQb;
    cout << prg << endl;
    printSettings(settings);
    if(!argsOK(settings)) return EXIT_FAILURE;
    if(!settings.execute) {
      Result res = QSys::testProgram(settings.numQb, &prg, settings.cases, settings.finalQb, settings.threshold);
      cout << endl << res;
    }
    else {
      std::vector<QSys*> res = QSys::execute(settings.numQb, &prg, settings.oracleTT);
      for(int i=0; i<res.size(); i++) {
        cout << *res[i];
      }
      return EXIT_SUCCESS;
    }
  }
  catch(Error e) {
    cerr << "\n" << e.s;
    cerr << "\nProgram aborted.";
    return EXIT_FAILURE;
  }
  return EXIT_SUCCESS;
}

void getOptions(int argc, char **argv, Settings &set) {
  char buf[256];
  set.execute = false;
  set.filename = NULL;
  if(argc<2) printUsage();
  for(int i=1; i<argc-1; i++) {
    if(argv[i][0]!='-') printUsage();
    switch(argv[i][1]) {
      case 'c':
        for(int j=i+1;j<argc-1;j++) {
          if(argv[j][0]=='-') break;
          if(strlen(argv[j])<3) printUsage();
          int k;
          for(k=0; k<strlen(argv[j])-2; k++) {
            if(argv[j][k]=='-') break;
            if(argv[j][k]!='0' && argv[j][k]!='1') printUsage();
            buf[k] = argv[j][k];
          }
          buf[k] = '\0';
          if(argv[j][k]!='-') printUsage();
          char *endpnt;
          set.cases.push_back(TestCase(buf,strtol(argv[j]+k+1,&endpnt,10)));
          if(*endpnt!='\0') printUsage();
          i++;
        }
        break;
      case 'f':
        for(int j=i+1;j<argc-1;j++) {
          if(argv[j][0]=='-') break;
          char *endpnt;
          int qb = strtol(argv[j],&endpnt,10);
          if(*endpnt!='\0') printUsage();
          set.finalQb.append(qb);
          i++;
        }
        break;
      case 'n': {
        char *endpnt;
        int num = strtol(argv[i+1],&endpnt,10);
        if(*endpnt!='\0') printUsage();
        set.numQb = num;
        i++;
      } break;
      case 't': {
        char *endpnt;
        int thresh = strtol(argv[i+1],&endpnt,10);
        if(*endpnt!='\0') printUsage();
        set.threshold = thresh;
        i++;
      } break;
      case 'x': {
        set.execute = true;
        if(i>=argc-2) break;
        if(argv[i+1][0]=='-') break;
        if(strlen(argv[i+1])<1) printUsage();
        int k;
        for(k=0; k<strlen(argv[i+1]); k++) {
          if(argv[i+1][k]!='0' && argv[i+1][k]!='1') printUsage();
          buf[k] = argv[i+1][k];
        }
        buf[k] = '\0';
        set.oracleTT = string(buf);
        i++;
      } break;
      case 'h': printUsage(); break;
      case 'v': printVersion(); break;
      case '-': 
        if(!strcmp(argv[i],"--version")) printVersion();
        else if(!strcmp(argv[i],"--help")) printUsage();
        break;
    }
  }
  /* file argument */
  if(!strcmp(argv[argc-1],"--version") || !strcmp(argv[argc-1],"-v")) 
    printVersion(); 
  if(!strcmp(argv[argc-1],"--help") || !strcmp(argv[argc-1],"-h"))
    printUsage();
  set.filename = argv[argc-1];
}

void printUsage() {
  cout << endl;
  cout << "This is a command line client for QGame++, a Quantum Gate And Measurement Emulator.\n"
       << "qgame lets you test and execute QGAME programs and produce statistical data.\n\n"
       << "To simply execute the program and print the results, give -x, followed by an oracle\n"
       << "truth table if needed (an oracle truth table is given as a string of 0 and 1, which\n"
       << "represents the right-hand column of the truth table).\n"
       << "To run the program on a number of test cases and produce statistical data describing\n"
       << "the error probabilities and other quality measures, give -c followed by a list of\n"
       << "test cases. A test case is given as an oracle truth table, followed by a minus, followed\n"
       << "by the value one wants to measure across the final measurement qubits.\n\n"
       ;
       
  cout << "Usage: qgame [OPTIONS] FILE\n\n"
       << "Options:\n"
       << "  -c CASE1 CASE2... \tList of test cases\n"
       << "  -f QB1 QB2...     \tFinal measurement qubits\n" 
       << "  -h                \tThis information\n"
       << "  -n NUMBER         \tNumber of qubits to use in the system\n"
       << "  -t THRESHOLD      \tThreshold for error probability\n"
       << "  -v                \tVersion information\n"
       << "  -x [ORACLETT]     \tExecute program\n"
       << "Either -c or -x must be given.  Use '-' as FILE to get input from STDIN instead of a file.\n"
       << endl
       << "Examples:\n"
       << "  qgame -c 0000-0 0001-1 FILE      Run FILE with specified test cases\n"
       << "  qgame -x -n 3 FILE               Execute FILE and print resulting systems.\n"
       ;


  exit(EXIT_FAILURE);
}

void printVersion() {
  cout << endl;
  cout << "qgame 0.1\n\n"
       << "Copyright (C) 2003-2004 Manuel Nickschas\n"
       << "This is free software; see the source for copying conditions.  There is NO\n"
       << "warranty; not even for MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.\n\n"
       << "Written by Manuel Nickschas <qgame-devel@nickschas.de>.\n"
       << "The original version of QGAME was implemented in Common Lisp by Lee Spector.\n"
       ;
       
  exit(EXIT_FAILURE);
}

void printSettings(const Settings &set) {
  if(set.execute) cout << "\nExecuting";
  else cout << "\nTesting";
  cout << " program ";
  if(!strcmp(set.filename,"-")) cout << "<STDIN>";
  else cout << "\"" << set.filename << "\"";
  cout << ",\nusing a system of " << set.numQb << " qubits and a threshold of "<<set.threshold;
  if(set.execute) {
    cout << "\nwith ";
    if(set.oracleTT.length()>0) cout << "the oracle truth table [ " << set.oracleTT << " ].\n";
    else cout << " no oracle truth table.\n";
  }
  else {
    cout << "\non the following test cases:\n";
    for(int t=0; t<set.cases.size(); t++) {
      cout << "  " << set.cases[t].oracleTT << "-" << set.cases[t].desiredRes << "\n";
    }
  }
  cout << "Final measurement qubits are " << set.finalQb << ".\n\n";
}

QubitList getQubits(int num, istream &is) throw(Error) {
  vector<int> l;
  QubitList res;
  int n;
  if(num>0) {
    for(int i=0; i<num; i++) {
      is >> n;
      l.push_back(n);
    }
  } else {
    while(is.peek()!=')' && is >> n) l.push_back(n);
  }
  for(int i=l.size()-1; i>=0; i--) {
    res.append(l[i]);
  }
  return res;  
}

bool argsOK(const Settings &set) {
  if(set.execute) { // execute, no testing
    
  } else {
    if(set.cases.size()<1) {
      cerr << "\nYou must specify at least one test case for testing a program! [-c]\n";
      return false;
    }
  }
  if(set.numQb<1) {
    cerr << "\nYou have to specify the number of qubits to use! [-n]\n";
    return false;
  }
  if(!set.filename) {
    cerr << "\nYou have to specify a file name!\n";
    return false;
  }
  if(!set.finalQb.size()) {
    cerr << "\nYou must specify a list of final measurement qubits! [-f]\n";
    return false;
  }
  if(set.threshold<0) {
    cerr << "\nYou must specify an error threshold! [-t]\n";    
    return false;
  }
  return true;
}
